/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"
#include "io_read.h"
#include "io_write.h"
#include "version.h"

#define MAX_COUNT_OF_DEVICES 256

int gs_canceled = 0;
int gs_major = -1;
int gs_minor = -1;
int gs_io_dev_added = 0;
struct cdev gs_cdev;
struct class *gs_ipf_class = NULL;
struct device *gs_device = NULL;

DECLARE_WAIT_QUEUE_HEAD(read_wait);
//DECLARE_WAIT_QUEUE_HEAD(cancel_wait);

int g_device_opened = 0;

/**
 *	io_dev_iov_total
 */
#undef __FUNC__
#define __FUNC__ "io_dev_iov_total()"
static inline ssize_t
io_dev_iov_total(
	const struct iovec *iv,
	unsigned long count
	)
{
	unsigned long i = 0;
	ssize_t len = 0;

	for (i = 0, len = 0; i < count; i++) {
		len += iv[i].iov_len;
	}

	return len;
}

/**
 *	io_dev_aio_read
 */
#undef __FUNC__
#define __FUNC__ "io_dev_aio_read()"
static ssize_t 
io_dev_aio_read(
	struct kiocb *iocb,
	const struct iovec *iv,
	unsigned long count,
	loff_t pos
	)
{
	//struct file *file = iocb->ki_filp;
	//struct tun_struct *tun = file->private_data;	// private data		
	ssize_t len = 0;	

	if ((len = io_dev_iov_total(iv, count)) < 0) {	
		TRACE(LOG_WARN, "%s: io_dev_iov_total(): len=%d", __FUNC__, len);
		return -EINVAL;
	}

	return io_read_to_iovec(iocb->ki_filp, (struct iovec*)iv, len);
}

/**
 *	io_dev_aio_write
 */
#undef __FUNC__
#define __FUNC__ "io_dev_aio_write()"
static ssize_t 
io_dev_aio_write(
	struct kiocb *iocb,
	const struct iovec *iv,
	unsigned long count,
	loff_t pos
	)
{
	ssize_t len = 0;	

	if ((len = io_dev_iov_total(iv, count)) < 0) {	
		TRACE(LOG_WARN, "%s: io_dev_iov_total(): len=%d", __FUNC__, len);
		return -EINVAL;
	}

	return io_write_from_iovec(iocb->ki_filp, (struct iovec*)iv, len);
}

/**
 *	io_dev_wakeup
 */
#undef __FUNC__
#define __FUNC__ "io_dev_wakeup()"
void
io_dev_wakeup(
	void
	)
{
	TRACE(LOG_DBG, "%s: trying to wake-up pending io request.", __FUNC__);
		
	wake_up_interruptible(&read_wait);	
}

#if !defined (USE_BLOCKING_READ)

/**
 *	io_dev_poll (select)
 *
 *	POLLIN      Data is available on the file descriptor for reading.
 *
 *	POLLNORM    Same as POLLIN.
 *
 *  POLLPRI     Same as POLLIN.
 *
 *  POLLOUT     Data can be written to the file descriptor without blocking.
 *
 *  POLLERR     This flag is not used in this implementation and is provided
 *              only for source code compatibility.
 *
 *  POLLHUP     The file descriptor was valid before the polling process and
 *              invalid after.  Presumably, this means that the file descrip-
 *              tor was closed sometime during the poll.
 *
 *  POLLNVAL    The corresponding file descriptor is invalid.
 *
 *  POLLRDNORM  Same as POLLIN.
 *
 *  POLLRDBAND  Same as POLLIN.
 *
 *  POLLWRNORM  Same as POLLOUT.
 *
 *  POLLWRBAND  Same as POLLOUT.
 *
 *  POLLMSG     This flag is not used in this implementation and is provided
 *              only for source code compatibility.
 *
 */
#undef __FUNC__
#define __FUNC__ "io_dev_poll()"
static unsigned int 
io_dev_poll(
	struct file *filp,
	poll_table *wait
	)
{	
	unsigned int mask = POLLPRI;			

	if (gs_canceled) {
		TRACE(LOG_INFO, "%s: io request has been canceled.", __FUNC__);		
		//gs_canceled = 0;
		//wake_up_interruptible(&cancel_wait);
		return mask;
	}

	do {

		//
		//	We are always writable
		//
		mask = POLLOUT | POLLWRNORM;

		//
		//	only insert into poll table and immediately return
		//
		poll_wait(filp, &read_wait, wait);
		
		//
		//	empty queue ?
		//						
		if (!(queue_is_empty(&g_main_queue))) {
			mask |= POLLIN | POLLRDNORM;
		}		

	} while(0);

	TRACE(LOG_DBG, "%s: filp=%p, wait=%p, writable=%s, readable=%s",
		__FUNC__, 
		filp,
		wait,
		((mask & POLLOUT) ? ("yes") : ("no")),
		((mask & POLLIN) ? ("yes") : ("no"))
		);

	return mask;
}

#endif // USE_BLOCKING_READ

/**
 *	io_dev_ioctl
 */
#undef __FUNC__
#define __FUNC__ "io_dev_ioctl()"
int 
io_dev_ioctl(
	struct inode *inode,
	struct file *file,
	unsigned int cmd,
	unsigned long arg
	)
{	
	int err = -EINVAL;	

	switch (cmd) {

		case KIPF_IOCTL_GET_INFO:
			{
				char version[KIPF_VERSION_STRING_LEN];
				struct kipf_info *info = (struct kipf_info __user *) arg;

				TRACE(LOG_DBG, "%s: KIPF_IOCTL_GET_INFO", __FUNC__);

				if (put_user(g_linipf_cfg.max_packet_size, &info->max_rw_buffer_size)) {
					TRACE(LOG_WARN, "%s: KIPF_IOCTL_GET_INFO: put_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}

				if (put_user(KIPF_MAJOR_VERSION, &info->api_version.major)) {
					TRACE(LOG_WARN, "%s: KIPF_IOCTL_GET_INFO: put_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}

				if (put_user(KIPF_MINOR_VERSION, &info->api_version.minor)) {
					TRACE(LOG_WARN, "%s: KIPF_IOCTL_GET_INFO: put_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}

				memset(version, 0, sizeof(version));
				snprintf(version, sizeof(version) - 1, "%d.%d.%d.%d", MAJOR_PRODUCT_VERSION, MINOR_PRODUCT_VERSION, REVISION_NUMBER, BUILD_NUMBER);

				if (copy_to_user(&info->version, version, MIN(sizeof(info->version) - 1, strlen(version) + 1))) {
					TRACE(LOG_WARN, "%s: KIPF_IOCTL_GET_INFO: copy_to_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}

				err = 0;
			}
			break;

		case KIPF_IOCTL_SET_DEFAULT_TRAFFIC_RULE:
			{								
				unsigned int rule = 0;

				TRACE(LOG_DBG, "%s: KIPF_IOCTL_SET_DEFAULT_TRAFFIC_RULE", __FUNC__);

				if (get_user(rule, (unsigned int __user *) arg)) {
					TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}								
		
				g_linipf_cfg.block_traffic_when_api_is_closed = ((rule == TRAFFIC_BLOCKED) ? (1) : (0));

				TRACE(LOG_DBG, "%s: block_traffic_when_api_is_closed: %d", __FUNC__, g_linipf_cfg.block_traffic_when_api_is_closed);

				if (g_linipf_cfg.block_traffic_when_api_is_closed) {
					g_linipf_cfg.start_time = 0xffffffff - START_TIMEOUT;
					g_linipf_cfg.timeout_expired = 0;
				} else {
					g_linipf_cfg.start_time = 0;
					g_linipf_cfg.timeout_expired = 1;
				}

				err = 0;
			}
			break;

		case KIPF_IOCTL_CANCEL_PENDING_READ:
			{
				//int ret = 0;

				TRACE(LOG_INFO, "%s: KIPF_IOCTL_CANCEL_PENDING_READ", __FUNC__);

				//
				//	mark device as canceled
				//
				gs_canceled = 1;

				io_dev_wakeup();

				//ret = wait_event_interruptible_timeout(cancel_wait, (gs_canceled == 0), msecs_to_jiffies(10000));
				//TRACE(LOG_INFO, "%s: wait_event_interruptible_timeout: %d", __FUNC__, ret);

				err = 0;
			}
			break;

		case KIPF_IOCTL_HANDLE_PACKET:
			{
				IOCTL_REQ ioctlRequest;				

				if (copy_from_user(&ioctlRequest, (IOCTL_REQ __user *) arg, sizeof(IOCTL_REQ))) {
					TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __FUNC__);
					err = -EFAULT;
					break;
				}		

				if (ioctlRequest.action.type == KIPF_ACTION_SEND_OUT || ioctlRequest.action.type == KIPF_ACTION_SEND_IN) {
					err = io_write_handle_data(&ioctlRequest.action, ioctlRequest.pkt, ioctlRequest.pkt_len);
				} else {
					err = io_write_handle_packet(&ioctlRequest.action, (struct kipf_packet*)ioctlRequest.packet);
				}
			}
			break;

		default:
			{
				TRACE(LOG_WARN, "%s: unknown ioctl cmd=0x%08X", __FUNC__, cmd);
			}
			break;
	}

	return err;
}

/**
 *	io_dev_release
 */
#undef __FUNC__
#define __FUNC__ "io_dev_release()"
static int 
io_dev_release(
	struct inode *inode,
	struct file *filp
	)
{	
	if (g_device_opened) {
		struct sk_buff *skb = NULL;			

		//
		//	Disable ip_filter_hook() queueing
		//
		g_device_opened = 0;

		//
		//	Disable packet queueing
		//
		queue_enable(&g_main_queue, 0);
		queue_enable(&g_pending_queue, 0);
	
		//
		//	Flush pending skb's
		//
		while ((skb = queue_remove(&g_main_queue, NULL, NULL))) {		
			ip_filter_issue_verdict(skb, NF_DROP);
		}	

		while ((skb = queue_remove(&g_pending_queue, NULL, NULL))) {		
			ip_filter_issue_verdict(skb, NF_DROP);
		}	

		if (IS_PTR_VALID(filp->private_data)) {
			filp->private_data = NULL;		
		}		

		TRACE(LOG_INFO, "%s: char device \"/dev/kipf\" closed.", __FUNC__);
	}	

	return 0;
}

/**
 *	io_dev_open
 */
#undef __FUNC__
#define __FUNC__ "io_dev_open()"
static int 
io_dev_open(
	struct inode *inode,
	struct file *filp
	)
{						
	//
	//	already opened ?
	//
	if (g_device_opened) {
		TRACE(LOG_INFO, "%s: char device \"/dev/kipf\" already opened.", __FUNC__);			
		return -EACCES;
	}			

	//
	//	ToDo: Add some usefull value
	//
	filp->private_data = (void*) 0x12345678;
		
	//
	//	Enable packet queueing
	//
	queue_enable(&g_main_queue, 1);
	queue_enable(&g_pending_queue, 1);	

	//
	//	Signal all up net devices
	//
	netdev_signal_all_up();

	//
	//	Enable ip_filter_hook() queueing
	//
	g_device_opened = 1;

	gs_canceled = 0;

	TRACE(LOG_INFO, "%s: char device \"/dev/kipf\" opened.", __FUNC__);

	return 0;
}

//
//	vpn_fops
//
static const struct file_operations vpn_fops = {	
	.owner		= THIS_MODULE,
#if !defined (USE_BLOCKING_READ)
	.poll		= io_dev_poll,
#endif
	.ioctl		= io_dev_ioctl,	
	.open		= io_dev_open,
	.release	= io_dev_release,
	.read		= do_sync_read,
	.aio_read	= io_dev_aio_read,
	.write		= do_sync_write,
	.aio_write	= io_dev_aio_write
};

/**
 *	io_dev_init
 */
#undef __FUNC__
#define __FUNC__ "io_dev_init()"
int
io_dev_init(
	void
	)
{
	dev_t dev;	
	int err = 0;        		

	do {

		gs_ipf_class = ERR_PTR(-ENOMEM);
		gs_device = ERR_PTR(-ENOMEM);
	
		if((err = alloc_chrdev_region(&dev, 0, MAX_COUNT_OF_DEVICES, "kipf"))) {		
			TRACE(LOG_ERR, "%s: alloc_chrdev_region(): err=%d", __FUNC__, err);
			break;
		}

		gs_major = MAJOR(dev);
		gs_minor = MINOR(dev);	

		cdev_init(&gs_cdev, &vpn_fops);		 

		gs_cdev.owner = THIS_MODULE;

		if ((err = cdev_add(&gs_cdev, dev, MAX_COUNT_OF_DEVICES))) {		
			TRACE(LOG_ERR, "%s: cdev_add(): err=%d", __FUNC__, err);
			break;
		}

		gs_io_dev_added = 1;
		
		if (IS_ERR((gs_ipf_class = class_create(THIS_MODULE, "kipf")))) {		
			err = PTR_ERR(gs_ipf_class);		
			TRACE(LOG_ERR, "%s: class_create(): err=%d", __FUNC__, err);		
			break;
		}		

#if defined(LINUX_KERNEL_2_6_22_TILL_2_6_26)
		if (IS_ERR((gs_device = device_create(gs_ipf_class, NULL, dev, "kipf")))) {
#elif defined(LINUX_KERNEL_2_6_27)
		if (IS_ERR((gs_device = device_create_drvdata(gs_ipf_class, NULL, dev, NULL, "kipf")))) {
#else
		if (IS_ERR((gs_device = device_create(gs_ipf_class, NULL, dev, NULL, "kipf")))) {
#endif
			err = PTR_ERR(gs_device);							
			TRACE(LOG_ERR, "%s: device_create(): err=%d", __FUNC__, err);		
			break;
		}

		TRACE(LOG_INFO, "%s: char device \"/dev/kipf\" created. (major=%d, minor=%d)", __FUNC__, gs_major, gs_minor);

	} while(0);	

	return err;
}

/**
 *	io_dev_close
 */
#undef __FUNC__
#define __FUNC__ "io_dev_close()"
int
io_dev_close(
	void
	)
{	
	if (gs_major != -1) {

		if (gs_io_dev_added) {

			if (!IS_ERR(gs_ipf_class)) {

				if (!IS_ERR(gs_device)) {			
					device_destroy(gs_ipf_class, MKDEV(gs_major, gs_minor));		
					gs_device = ERR_PTR(-ENOMEM);
				}		

				class_destroy(gs_ipf_class);
				gs_ipf_class = ERR_PTR(-ENOMEM);
			}

			cdev_del(&gs_cdev);
			gs_io_dev_added = 0;
		}

		unregister_chrdev_region(MKDEV(gs_major, gs_minor), MAX_COUNT_OF_DEVICES);

		gs_major = -1;
		gs_minor = -1;

		TRACE(LOG_INFO, "%s: char device \"/dev/kipf\" removed.", __FUNC__);
	}

	return 0;
}

