/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"

/**
 *	struct grev1hdr
 */
struct grev1hdr {
	u_int8_t recur:3;
	u_int8_t is_strict:1;
	u_int8_t has_seq:1;
	u_int8_t has_key:1;
	u_int8_t has_routing:1;
	u_int8_t has_cksum:1;
	u_int8_t version:3;
	u_int8_t flags:4;
	u_int8_t has_ack:1;
	u_int16_t protocol;
	u_int16_t plen;
	u_int16_t callid;
	u_int32_t seq_nr;
	u_int32_t ack_nr;
};

/*
 *	Macro for translating hw address to interface's id (uses in libkwfapi and driver)
 */
#define hwAddressToId(id, hwAddress)					\
	sprintf((id), "/device/eth/{%02x-%02x-%02x-%02x-%02x-%02x}",	\
		(hwAddress)[0],						\
		(hwAddress)[1],						\
		(hwAddress)[2],						\
		(hwAddress)[3],						\
		(hwAddress)[4],						\
		(hwAddress)[5]						\
		);							\

/*
 *	Macro for translating RAS name (pppX) to interface's id (same principle used in RasControl in engine)
 */
#define ifaceNameToId(id, name)			\
	sprintf((id), "/device/ppp/%s",		\
		(name) + 3			\
		);		

#define MIN_BYTES_TO_COPY 128

/**
 *	io_read_get_ip_cp_len - it can return max 128 (IPinIP or IP+TCP or MIN_BYTES_TO_COPY)
 */
#undef __FUNC__
#define __FUNC__ "io_read_get_ip_cp_len()"
static u_int32_t
io_read_get_ip_cp_len(	
	struct iphdr *ip,
	u_int32_t len
	)
{
	u_int32_t ip_copied_len = 0;
	void *hdr = NULL;		

	//
	//	IP header (max. 64) ihl = 4bits
	//
	ip_copied_len = (ip->ihl * 4);

	//
	// Invalid length of IP ?
	//
	if (ip_copied_len < 20) {
		return ip_copied_len;
	}	

	hdr = ((u_int8_t*)ip + ip_copied_len);

	switch (ip->protocol) {

		//
		//	TCP header (max. 64) data_offs = 4bits
		//
		case IPPROTO_TCP:
			{				
				struct tcphdr* tcp = hdr;

				ip_copied_len += (tcp->doff * 4);
			}
			break;

		//
		//	UDP header
		//
		case IPPROTO_UDP:
			ip_copied_len += sizeof(struct udphdr);
			break;

		//
		//	GRE header
		//
		case IPPROTO_GRE:
			ip_copied_len += sizeof(struct grev1hdr);
			break;

		//
		//	"IPinIP" header
		//
		case IPPROTO_IPIP:
			{
				struct iphdr *ip2 = hdr;

				ip_copied_len += (ip2->ihl * 4);
			}
			break;

		//
		//	ICMP header
		//
		case IPPROTO_ICMP:
			{
				struct icmphdr *icmp = hdr;
				
				//
				//	ToDo: add only specific count of bytes ?
				//
				switch (icmp->type) {
				}
			}	

		default:
			ip_copied_len = MIN_BYTES_TO_COPY;
			break;
	}

	if (ip_copied_len > len) {
		ip_copied_len = len;	
	}	

	return ip_copied_len;
}

/**
 *	io_read_net_skb
 */
#undef __FUNC__
#define __FUNC__ "io_read_net_skb()"
static int
io_read_net_skb(
	struct kipf_packet *packet,
	struct sk_buff *skb
	)
{		
	u_int8_t *dst = NULL;	
	int bytes_up = 0;
#if defined (USE_SCATTER_GATHER)
	u_int8_t kbuffer[MIN_BYTES_TO_COPY];
	int err = 0;
#endif

	bool mac_is_set = ((skb_mac_header_was_set(skb) && skb->mac_len) ? (true) : (false));

	int ip_copied_len = io_read_get_ip_cp_len((struct iphdr*)ip_hdr(skb), skb->len);		

	if (ip_copied_len < 20) {
		return -EINVAL;
	}

	if (put_user(KIPF_MAGIC(KIPF_MAGIC_PACKET), &packet->header.magic)) {
		return -EFAULT;
	}

	if (put_user((u_int32_t) skb->dev->ifindex, &packet->ifindex)) {
		return -EFAULT;
	}

	if (put_user((u_int32_t) skb, &packet->id)) {
		return -EFAULT;
	}

	if (put_user((((GET_CB(skb))->received) ? (1) : (0)), &packet->received)) {
		return -EFAULT;
	}

	if (put_user(sizeof(struct ethhdr) + skb->len, &packet->len)) {
		return -EFAULT;
	}
		
	dst = packet->data;

	//
	//	Valid MAC ?
	//
	if (mac_is_set) {		

		if (copy_to_user(dst, skb_mac_header(skb), skb->mac_len)) {
			return -EFAULT;
		}

	} else {		
		u_int8_t null_addr[ETH_ALEN] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
		struct ethhdr *eh = (struct ethhdr*) dst;		

		//
		//	Fill fake mac header
		//  
		if ((GET_CB(skb))->received) {
			if (copy_to_user(eh->h_dest, skb->dev->dev_addr, ETH_ALEN)) {
				return -EFAULT;
			}
			if (copy_to_user(eh->h_source, null_addr, ETH_ALEN)) {
				return -EFAULT;
			}
		} else {
			struct dst_entry *dst = skb->dst;

			if (copy_to_user(eh->h_dest, null_addr, ETH_ALEN)) {
				return -EFAULT;
			}

			if (copy_to_user(eh->h_source, skb->dev->dev_addr, ETH_ALEN)) {
				return -EFAULT;
			}

			if (dst) {
				if (dst->hh) {
					if (dst->hh->hh_type == __constant_htons(ETH_P_IP)) {
						struct ethhdr *_eh = (struct ethhdr*) (((u8*) dst->hh->hh_data) + (HH_DATA_OFF(sizeof(*_eh))));
						if (copy_to_user(eh->h_dest, _eh->h_dest, ETH_ALEN)) {
							return -EFAULT;
						}
					}
				} else if (dst->neighbour) {
					if (copy_to_user(eh->h_dest, dst->neighbour->ha, MIN(sizeof(dst->neighbour->ha), ETH_ALEN))) {
						return -EFAULT;
					}
				}
			}			
		}

		if (put_user(htons(ETH_P_IP), &eh->h_proto)) {
			return -EFAULT;
		}		
	}				

	dst += sizeof(struct ethhdr);

#if defined (USE_SCATTER_GATHER)

	bytes_up = MIN(ip_copied_len, sizeof(kbuffer));

	//
	//	copy from skb into kernel memory first
	//
	if ((err = skb_copy_bits(skb, 0, &kbuffer, bytes_up))) {			
		TRACE(LOG_WARN, "%s: skb_copy_bits(): skb=%p, bytes_up=%d, err=%d", __FUNC__, skb, bytes_up, err);
		return -EINVAL;
	}

	//	
	//	and then from kernel memory into user space buffer
	//
	if (copy_to_user(dst, &kbuffer, bytes_up)) {
		return -EFAULT;
	}

#else  // USE_SCATTER_GATHER

	bytes_up = ip_frag_copy_packet_data(skb, dst, ip_copied_len, 0);		

#endif // USE_SCATTER_GATHER

	(GET_CB(skb))->copied_len = sizeof(struct ethhdr) + bytes_up;

	if (put_user((GET_CB(skb))->copied_len, &packet->data_len)) {
		return -EFAULT;
	}	

	TRACE(LOG_DBG, "%s: skb=%p, mac_is_set=%s, mac_len=%d, ip_copied_len=%d, bytes_up=%d, copied_len=%d",
		__FUNC__, skb, ((skb_mac_header_was_set(skb)) ? ("yes") : ("no")),
		skb->mac_len, ip_copied_len, bytes_up, (GET_CB(skb))->copied_len);

	return 0;
}

/**
 *	io_read_dev_skb
 */
#undef __FUNC__
#define __FUNC__ "io_read_dev_skb()"
static int
io_read_dev_skb(
	struct kipf_iface *iface,
	struct sk_buff *skb
	)
{
	char tmp[KIPF_MAX_NSIZE];		

	//
	//	NULL packet ?
	//
	if (skb->dev == NULL) {

		if (put_user(KIPF_MAGIC(KIPF_MAGIC_IFACE), &iface->header.magic)) {
			return -EFAULT;
		}

		if (put_user(0, &iface->ifindex)) {
			return -EFAULT;
		}

		return 0;
	}

	if (put_user(KIPF_MAGIC(KIPF_MAGIC_IFACE), &iface->header.magic)) {
		return -EFAULT;
	}

	if (put_user(((skb->dev->flags & IFF_UP) ? (KIPF_IFACE_UP) : (0)), &iface->flags)) {
		return -EFAULT;
	}

	//
	//	little hack: master serves as storage for original dev (adapter id)
	//
	//if (put_user((u_int32_t) skb->dev->master, &iface->ifindex)) {
	//	return -EFAULT;
	//}

	if (put_user(skb->dev->mtu, &iface->mtu)) {
		return -EFAULT;
	}

	if (put_user(((skb->dev->type == ARPHRD_ETHER) ? (KIPF_HW_TYPE_ETH) : ((skb->dev->type == ARPHRD_PPP) ? (KIPF_HW_TYPE_PPP) : (KIPF_HW_TYPE_UNK))), &iface->hw_type)) {
		return -EFAULT;
	}

	if (put_user(skb->dev->ifindex, &iface->ifindex)) {
		return -EFAULT;
	}

	if (copy_to_user(iface->hw_addr, skb->dev->dev_addr, MIN(sizeof(iface->hw_addr), sizeof(skb->dev->dev_addr)))) {
		return -EFAULT;
	}

	//
	//	create id
	//
	if (skb->dev->type == ARPHRD_PPP) {
		if (strlen(skb->dev->name) < 4)
			return -EFAULT;
		ifaceNameToId(tmp, skb->dev->name);	
	}
	else
		hwAddressToId(tmp, skb->dev->dev_addr);	

	if (copy_to_user(iface->id, tmp, MIN(sizeof(iface->id), strlen(tmp) + 1))) {
		return -EFAULT;			
	}

	if (copy_to_user(iface->name, skb->dev->name, MIN(sizeof(iface->name), strlen(skb->dev->name) + 1))) {
		return -EFAULT;
	}

	return 0;
}

/**
 *	io_read_to_iovec
 */
#undef __FUNC__
#define __FUNC__ "io_read_to_iovec()"
ssize_t
io_read_to_iovec(			
	struct file *file,
	struct iovec *iv,
	ssize_t len
	)
{	
#if defined (USE_BLOCKING_READ)
	DECLARE_WAITQUEUE(wait, current);
#endif
	struct sk_buff *skb = NULL;	
	ssize_t readed_skb = 0;
	int err = 0;
	u_int32_t i = 0;		
	READ_REQ readRequest;

	if (len != sizeof(READ_REQ)) {
		TRACE(LOG_WARN, "%s: invalid length. (%d)", __FUNC__, len);
		return -EINVAL;
	}

	if (memcpy_fromiovec((unsigned char*) &readRequest, iv, len)) {
		TRACE(LOG_WARN, "%s: memcpy_fromiovec(): segmentation fault.", __FUNC__);		
		return -EFAULT;
	}
	
#if defined (USE_BLOCKING_READ)
	add_wait_queue(&read_wait, &wait);
	set_current_state(TASK_INTERRUPTIBLE);
#endif // USE_BLOCKING_READ

	for (i = 0; i < readRequest.count; i++) {		

#if defined (USE_BLOCKING_READ)
try_again:
#endif

		skb = queue_remove(&g_main_queue, NULL, NULL);					

		if (!IS_PTR_VALID(skb)) {

			if (readed_skb == 0) {	
				
#if defined (USE_BLOCKING_READ)

				TRACE(LOG_DBG, "%s: schedule(): going to sleep...", __FUNC__);
			
				//
				//	Nothing to read, let's sleep
				//				
				schedule();				

				//
				//	After you slept you should check if a signal occurred: the Unix/Linux way of handling signals
				//	is to temporarily exit the system call with the -ERESTARTSYS error. The system call entry code
				//	will switch back to user context, process the signal handler and then your system call will be
				//	restarted.
				//
				//	http://mail.nl.linux.org/kernelnewbies/2003-02/msg00305.html
				//
				if (signal_pending(current)) {
					TRACE(LOG_WARN, "%s: some system signal has occured.", __FUNC__);
					err = -ERESTARTSYS;
					//err = -EINTR;
					break;
				}

				goto try_again;

#else

				//
				//	file is NOT set as nonblocking ?
				//
				if (!(file->f_flags & O_NONBLOCK)) {
					TRACE(LOG_WARN, "%s: this module is compiled with nonblocking file support only.", __FUNC__);
					err = -EINVAL;
					break;
				}

				TRACE(LOG_DBG, "%s: no skb in the queue and file is set to nonblocking mode.", __FUNC__);
				err = -EAGAIN;

#endif // USE_BLOCKING_READ

			}

			break;
		}

		//
		//	Now we have a skb ...
		//
		TRACE(LOG_DBG, "%s: reading data from skb=%p, len=%d, dev='%s'", __FUNC__, skb, skb->len, ((skb->dev) ? (skb->dev->name) : ("???")));				

		//
		//	NETWORK_PACKET
		//
		if ((GET_CB(skb))->type == NETWORK_PACKET) {

			if ((err = io_read_net_skb((struct kipf_packet*) readRequest.headers[i], skb))) {
				TRACE(LOG_WARN, "%s: io_read_net_skb(): skb=%p, err=%d", __FUNC__, skb, err);
				ip_filter_issue_verdict(skb, NF_DROP);
				break;
			}			

			//
			//	Insert packet into pending queue
			//
			if ((err = queue_insert(&g_pending_queue, skb, 0))) {
				TRACE(LOG_WARN, "%s: queue_insert(): can't insert skb=%p into pending queue. err=%d", __FUNC__, skb, err);
				ip_filter_issue_verdict(skb, NF_DROP);
				break;
			}

			readed_skb++;

			continue;
		}

		//
		//	NET_DEV_PACKET
		//		
		if ((GET_CB(skb))->type == NET_DEV_PACKET) {

			if ((err = io_read_dev_skb((struct kipf_iface*) readRequest.headers[i], skb))) {
				TRACE(LOG_WARN, "%s: io_read_dev_skb(): skb=%p, err=%d", __FUNC__, skb, err);
				ip_filter_issue_verdict(skb, NF_DROP);
				break;
			}

			ip_filter_issue_verdict(skb, NF_DROP);

			readed_skb++;

			continue;
		} 			

		ip_filter_issue_verdict(skb, NF_DROP);
	}

#if defined (USE_BLOCKING_READ)
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&read_wait, &wait);		
#endif // USE_BLOCKING_READ	

	readRequest.count = err ? 0 : readed_skb;

	return ((err) ? (err) : (readed_skb));
}

