/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"
#include "socket.h"
#include "pkt.h"

static int pre_routing_hook_registered = 0;
static int post_routing_hook_registered = 0;
static int queue_handler_registered = 0;

struct ipf_queue g_main_queue;
struct ipf_queue g_pending_queue;

/**
 *	ip_filter_issue_verdict
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_issue_verdict()"
int
ip_filter_issue_verdict(
	struct sk_buff *skb,
	int verdict	
	)
{
	bool policy_routing = false;
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	struct nf_info *nf_data = NULL;
#else
	struct nf_queue_entry *nf_data = NULL;
#endif		

	if (!IS_PTR_VALID(skb)) {
		TRACE(LOG_WARN, "%s: skb is NULL ???", __FUNC__);
		return -EINVAL;
	}

	if ((GET_CB(skb))->type == NET_DEV_PACKET) {
		TRACE(LOG_DBG, "%s: skb=%p describing net device freed.", __FUNC__, skb);
		if (skb->dev) {
			free_nonpaged_mem(skb->dev, sizeof(struct net_device));
			skb->dev = NULL;
		}
		kfree_skb(skb);
		return 0;
	} 

	if ((GET_CB(skb))->type != NETWORK_PACKET) {
		TRACE(LOG_WARN, "%s: skb=%p with unknown type=0x%08X freed.", __FUNC__, skb, (GET_CB(skb))->type);
		kfree_skb(skb);
		return 0;
	} 	
	
	if ((policy_routing = (GET_CB(skb))->policy_routing)) {
		verdict = NF_STOLEN;
	}

	nf_data = (GET_CB(skb))->nf_data;

	if (!IS_PTR_VALID(nf_data)) {
		TRACE(LOG_WARN, "%s: skb=%p with NULL nf_data pointer freed.", __FUNC__, skb);
		kfree_skb(skb);
		return -EINVAL;
	}

	TRACE(LOG_DBG, "%s: skb=%p, len=%d, dev='%s', verdict='%s'",
		__FUNC__, skb, skb->len, ((skb->dev) ? (skb->dev->name) : ("???")),		
		((verdict == NF_DROP) ? ("dropped") :
		(verdict == NF_ACCEPT) ? ("accepted") :
		(verdict == NF_STOLEN) ? ("stolen") :
		(verdict == NF_QUEUE) ? ("queued") :
		(verdict == NF_REPEAT) ? ("repeated") : ("???")));

	/* clear only portion affected by our code */
	memset(GET_CB(skb), 0, sizeof(struct ipf_skb_cb));

	/* TCP input path (and probably other bits) assume to be called
	 * from softirq context, not from syscall, like ip_filter_issue_verdict is
	 * called.  TCP input path deadlocks with locks taken from timer
	 * softirq, e.g.  We therefore emulate this by local_bh_disable() */
	
	local_bh_disable();

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	nf_reinject(skb, nf_data, verdict);
#else
	nf_reinject(nf_data, verdict);
#endif
	
	local_bh_enable();

	if (policy_routing) {
		int ret = 0;

		TRACE(LOG_DBG, "%s: policy routing applied on skb=%p", __FUNC__, skb);

		/* it have to be called manually because some skb can be previously received (okfn is ip_rcv_finish) */
		if ((ret = ip_finish_output(skb))) {	
			TRACE(LOG_WARN, "%s: ip_finish_output(): %d", __FUNC__, ret);		
			return ret;
		}	
	}

	return 0;
}

/**
 *	ip_filter_check_timeout
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_check_timeout"
static u_int8_t
ip_filter_check_timeout(
	void
	)
{	
	unsigned long t = get_seconds();

	if ((g_linipf_cfg.start_time + START_TIMEOUT) < t) {

		g_linipf_cfg.timeout_expired = 1;

		TRACE(LOG_INFO, "%s: start timeout expired at %u", __FUNC__, t);			

		return 1;
	}

	return 0;
}

/**
 *	ip_filter_pre_process_packet
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_pre_process_packet"
static u_int32_t
ip_filter_pre_process_packet(
	void
	)
{
	if (!g_linipf_cfg.block_traffic_when_api_is_closed) {
		return NF_ACCEPT;
	}

	if (g_linipf_cfg.timeout_expired) {
		return NF_ACCEPT;
	}

	if (ip_filter_check_timeout()) {
		return NF_ACCEPT;
	}

	return NF_DROP;
}

/**
 *	ip_filter_hook
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_hook()"
static unsigned int 
ip_filter_hook(
	unsigned int hooknum,
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	struct sk_buff **pskb,
#else
	struct sk_buff *skb,
#endif
	const struct net_device *in,
	const struct net_device *out,
	int (*okfn)(struct sk_buff *)
	)
{		
	const struct net_device *dev = ((hooknum == NF_IP_PRE_ROUTING) ? (in) : (out));		
#if !defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	struct sk_buff **pskb = &skb;
#endif

	TRACE(LOG_DBG, "%s: hooknum=%d, skb=%p, dev='%s'",
		__FUNC__, hooknum, ((pskb) ? (*pskb) : (NULL)), ((dev) ? (dev->name) : ("???")));

	if (!pskb || !(*pskb) || !dev) {
		TRACE(LOG_DBG, "%s: invalid skb or dev ???", __FUNC__);	
		return NF_DROP;
	}	

	//
	//	packet was indicated via loopback device ?
	//
	if (dev->flags & IFF_LOOPBACK) {	

		if (IS_DEBUG_FLAG_SET()) {

			TRACE(LOG_DBG, "%s: packet handling on loopback device isn't permitted. skb=%p", __FUNC__, *pskb);	

			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));						
		}

		return NF_ACCEPT;
	}

	//
	//	is our char device opened ?
	//
	if (!g_device_opened) {	

		if (IS_DEBUG_FLAG_SET()) {
			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));	
		}

		return ip_filter_pre_process_packet();			
	}
	
	//
	//	Check if packet was previously indicated by pkt_send_in() ?
	//		
	if (hooknum == NF_IP_PRE_ROUTING 
		&&
		(*pskb)->mark == PKT_SOCKET_TAG) 
	{
		if (IS_DEBUG_FLAG_SET()) {

			TRACE(LOG_DBG, "%s: skb=%p sent by pkt_send_in().", __FUNC__, *pskb);

			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));						
		}
			
		(*pskb)->mark = 0;

		return NF_ACCEPT;
	}		

	//
	//	Check if packet was sent by pkt_send_out()
	//					
	/*
	if (hooknum == NF_IP_POST_ROUTING
		&&
		IS_PTR_VALID((*pskb)->sk)	//	can be NULL
		&&
		socket_is_our((*pskb)->sk->sk_socket))
	{		
		if (IS_DEBUG_FLAG_SET()) {

			TRACE(LOG_DBG, "%s: skb=%p sent by pkt_send_out().", __FUNC__, *pskb);	

			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));						
		}
		
		return NF_ACCEPT;
	}
	*/

	//
	//	Invalidate hardware checksum when packet is to be mangled, and
	//	complete checksum manually on outgoing path.
	//
	if ((*pskb)->ip_summed == CHECKSUM_PARTIAL
		||
		(*pskb)->ip_summed == CHECKSUM_COMPLETE)
	{
		int err = skb_checksum_help(*pskb);

		if (err) {

			TRACE(LOG_WARN, "%s: skb_checksum_help(): can't invalidate checksum on skb=%p, ip_summed=%d, err=%d",
				__FUNC__, *pskb, (*pskb)->ip_summed, err);

			if (IS_DEBUG_FLAG_SET()) {
				dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));
			}

			return NF_DROP;
		}
	}

	//
	//	Gather fragments (if any)
	//		
	*pskb = ip_frag_gather(*pskb);

	if (!IS_PTR_VALID(*pskb)) {		
		return NF_STOLEN;
	}

	if ((*pskb)->len + sizeof(struct ethhdr) > g_linipf_cfg.max_packet_data_size) {

		TRACE(LOG_WARN, "%s: Warning: too long packet. skb=%p, len=%d, max_packet_data_size=%d",
			__FUNC__, *pskb, (*pskb)->len + sizeof(struct ethhdr), g_linipf_cfg.max_packet_data_size);

		if (IS_DEBUG_FLAG_SET()) {
			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));
		}

		return NF_DROP;
	}

	if (IS_PTR_VALID(skb_shinfo(*pskb)->frag_list)) {

		struct sk_buff *frag_skb = skb_shinfo(*pskb)->frag_list;

		if (IS_PTR_VALID(skb_shinfo(frag_skb)->frag_list)) {

			TRACE(LOG_WARN, "%s: Warning: unimplemented type of packet. skb=%p, len=%d, 1. frag_list=%p, 2. frag_list=%p",
				__FUNC__, *pskb, (*pskb)->len , frag_skb, skb_shinfo(frag_skb)->frag_list);

			if (IS_DEBUG_FLAG_SET()) {
				dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));
			}

			return NF_DROP;
		}
	}
	
#if !defined (USE_SCATTER_GATHER)

	if (skb_shinfo(*pskb)->nr_frags) {

		TRACE(LOG_WARN, "%s: Warning: scatter/gather is not supported. skb=%p, len=%d, nr_frags=%d",
			__FUNC__, *pskb, (*pskb)->len , skb_shinfo(*pskb)->nr_frags);

		if (IS_DEBUG_FLAG_SET()) {
			dump_ip_packet((u_int8_t*) ip_hdr(*pskb), skb_headlen(*pskb));
		}

		return NF_DROP;
	}

#endif // USE_SCATTER_GATHER

	//
	//	Possible return values:
	//
	//	NF_ACCEPT: continue traversal as normal. 
	//	NF_DROP: drop the packet; don't continue traversal. 
	//	NF_STOLEN: I've taken over the packet; don't continue traversal. 
	//	NF_QUEUE: queue the packet (usually for userspace handling). 
	//	NF_REPEAT: call this hook again. 

	return NF_QUEUE;
}

//
//	ipt_ops
//
static struct nf_hook_ops ipt_ops[] = {
	{
		.hook		= ip_filter_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_PRE_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
	{
		.hook		= ip_filter_hook,
		.owner		= THIS_MODULE,
		.pf		= PF_INET,
		.hooknum	= NF_IP_POST_ROUTING,
		.priority	= NF_IP_PRI_FIRST,
	},
};

/**
 *	ip_filter_enqueue_handler
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_enqueue_handler()"
static int
ip_filter_enqueue_handler(
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	struct sk_buff *skb,
	struct nf_info *nf_data,
	unsigned int queuenum,
	void *data
#else
	struct nf_queue_entry *nf_data,
	unsigned int queuenum
#endif
	)
{				
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	unsigned int hooknum = nf_data->elem->hooknum;
#else
	void *data = NULL; // dummy value
	struct sk_buff *skb = nf_data->skb;
	unsigned int hooknum = nf_data->hook;
#endif

	if (hooknum == NF_IP_PRE_ROUTING) {	
		recv(skb, nf_data, data);
		return 0;
	} 
	
	if (hooknum == NF_IP_POST_ROUTING) {
		send(skb, nf_data, data);
		return 0;
	}

	TRACE(LOG_WARN, "%s: unknown hooknum=%d, skb=%p", __FUNC__, hooknum, skb);	

	return -EINVAL;
}

//
//	nfqh
//
static struct nf_queue_handler nfqh = {	
	.outfn	= &ip_filter_enqueue_handler,
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	.data   = (void*) 0x12345678,
#endif
	.name	= "IP filter",
};

/**
 *	ip_filter_init
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_init()"
int
ip_filter_init(
	void
	)
{
	int err = 0;

	queue_init(&g_main_queue);

	queue_init(&g_pending_queue);

	if ((err = nf_register_queue_handler(PF_INET, &nfqh))) {
		TRACE(LOG_ERR, "%s: nf_register_queue_handler(): err=%d", __FUNC__, err);
		return err;
	}

	TRACE(LOG_INFO, "%s: queue handler registered.", __FUNC__);

	queue_handler_registered = 1;
	
	if ((err = nf_register_hook(&ipt_ops[0]))) {
		TRACE(LOG_ERR, "%s: nf_register_hook(): hooknum=%d, err=%d", __FUNC__, ipt_ops[0].hooknum, err);
		return err;
	}

	TRACE(LOG_INFO, "%s: pre-routing hook registered. hooknum=%d", __FUNC__, ipt_ops[0].hooknum);

	pre_routing_hook_registered = 1;

	if ((err = nf_register_hook(&ipt_ops[1]))) {	
		TRACE(LOG_ERR, "%s: nf_register_hook(): hooknum=%d, err=%d", __FUNC__, ipt_ops[1].hooknum, err);
		return err;
	}

	TRACE(LOG_INFO, "%s: post-routing hook registered. hooknum=%d", __FUNC__, ipt_ops[1].hooknum);

	post_routing_hook_registered = 1;

	return err;
}

/**
 *	ip_filter_close
 */
#undef __FUNC__
#define __FUNC__ "ip_filter_close()"
int
ip_filter_close(
	void
	)
{	
	if (post_routing_hook_registered) {
		TRACE(LOG_INFO, "%s: post-routing hook unregistered.", __FUNC__);
		nf_unregister_hook(&ipt_ops[1]);
		post_routing_hook_registered = 0;
	}

	if (pre_routing_hook_registered) {
		TRACE(LOG_INFO, "%s: pre-routing hook unregistered.", __FUNC__);
		nf_unregister_hook(&ipt_ops[0]);
		pre_routing_hook_registered = 0;
	}

	if (queue_handler_registered) {
		TRACE(LOG_INFO, "%s: queue handler unregistered.", __FUNC__);
		nf_unregister_queue_handlers(&nfqh);
		queue_handler_registered = 0;
	}

	queue_close(&g_pending_queue);	

	queue_close(&g_main_queue);	

	return 0;
}

