/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"

/**
 *	netdev_event
 */
static int netdev_event(struct notifier_block *this, unsigned long event, void *ptr);

/**
 *	Locals
 */
static int gs_netdev_notifier_registered = 0;
static int gs_inet_notifier_registered = 0;

static struct notifier_block gs_ipf_netdev_notifier = {
	.notifier_call = netdev_event,
};
static struct notifier_block gs_ipf_inet_notifier = {
	.notifier_call	= netdev_event,
};

/**
 *	netdev_event
 */
#undef __FUNC__
#define __FUNC__ "netdev_event()"
static int 
netdev_event(
	struct notifier_block *this,
	unsigned long event,
	void *ptr
	)
{	
	struct net_device *dev = 
		((this == &gs_ipf_inet_notifier) ? (((struct in_ifaddr *)ptr)->ifa_dev->dev) : (ptr));	

	switch (event) {

		case NETDEV_UP:
			{
				TRACE(LOG_DBG, "%s: NETDEV_UP: '%s'", __FUNC__, dev->name);

				netdev_up(dev);
			}
			break;

		case NETDEV_DOWN:
			{
				TRACE(LOG_DBG, "%s: NETDEV_DOWN: '%s'", __FUNC__, dev->name);

				netdev_down(dev);
			}
			break;

		case NETDEV_REBOOT:
			TRACE(LOG_DBG, "%s: NETDEV_REBOOT: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_CHANGE:
			TRACE(LOG_DBG, "%s: NETDEV_CHANGE: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_REGISTER:
			TRACE(LOG_DBG, "%s: NETDEV_REGISTER: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_UNREGISTER:
			TRACE(LOG_DBG, "%s: NETDEV_UNREGISTER: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_CHANGEMTU:
			TRACE(LOG_DBG, "%s: NETDEV_CHANGEMTU: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_CHANGEADDR:
			TRACE(LOG_DBG, "%s: NETDEV_CHANGEADDR: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_GOING_DOWN:
			TRACE(LOG_DBG, "%s: NETDEV_GOING_DOWN: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_CHANGENAME:
			TRACE(LOG_DBG, "%s: NETDEV_CHANGENAME: '%s'", __FUNC__, dev->name);
			break;

		case NETDEV_FEAT_CHANGE:
			TRACE(LOG_DBG, "%s: NETDEV_FEAT_CHANGE: '%s'", __FUNC__, dev->name);
			break;

		default:
			TRACE(LOG_WARN, "%s: unknown net device event (%d). dev='%s'", __FUNC__, event, dev->name);
			break;
	}

	return NOTIFY_DONE;
}

/**
 *	netdev_dump_skb
 */
#undef __FUNC__
#define __FUNC__ "netdev_dump_skb()"
static void
netdev_dump_skb(
	struct sk_buff *skb,
	char *func_name
	)
{
	TRACE(LOG_DBG,
		"%s: '%s': skb=%p, dev=%p, mac=%02x:%02x:%02x:%02x:%02x:%02x, mtu=%d, state=%d, features=0x%X, ifindex=0x%X, "
		"flags=0x%X, type=%d, hard_header_len=%d, addr_len=%d",
		func_name,
		skb->dev->name,
		skb,
		skb->dev->master,
		skb->dev->dev_addr[0],
		skb->dev->dev_addr[1],
		skb->dev->dev_addr[2],
		skb->dev->dev_addr[3],
		skb->dev->dev_addr[4],
		skb->dev->dev_addr[5],			
		skb->dev->mtu,
		skb->dev->state,
		skb->dev->features,
		skb->dev->ifindex,			
		skb->dev->flags,
		skb->dev->type,
		skb->dev->hard_header_len,			
		skb->dev->addr_len					
		);
}

/**
 *	netdev_find_by_mac
 */
#undef __FUNC__
#define __FUNC__ "netdev_find_by_mac()"
struct net_device *
netdev_find_by_mac(
	unsigned char *mac,
	int len
	)
{
	struct net_device *dev = NULL;
	
	read_lock(&dev_base_lock);

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	for_each_netdev(dev) {	
#else
	for_each_netdev(&init_net, dev) {	
#endif
		if (!memcmp(dev->dev_addr, mac, MIN(sizeof(dev->dev_addr), len))) {
			read_unlock(&dev_base_lock);
			return dev;
		}
	}

	read_unlock(&dev_base_lock);

	return NULL;
}

/**
 *	netdev_find_by_ifindex
 */
#undef __FUNC__
#define __FUNC__ "netdev_find_by_ifindex()"
struct net_device *
netdev_find_by_ifindex(
	unsigned int ifindex	
	)
{
	struct net_device *dev = NULL;
	
	read_lock(&dev_base_lock);

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	for_each_netdev(dev) {	
#else
	for_each_netdev(&init_net, dev) {	
#endif
		if (dev->ifindex == ifindex) {
			read_unlock(&dev_base_lock);
			return dev;
		}
	}

	read_unlock(&dev_base_lock);

	return NULL;
}

/**
 *	netdev_enum_all - taken from linux-2.6.23/net/core/dev.c
 */
#undef __FUNC__
#define __FUNC__ "netdev_enum_all()"
void
netdev_enum_all(
	void
	)
{
	struct net_device *dev = NULL;
	
	read_lock(&dev_base_lock);
	
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	for_each_netdev(dev) {	
#else
	for_each_netdev(&init_net, dev) {	
#endif	
		TRACE(LOG_INFO,
			"%s: '%s': dev=%p, mac=%02x:%02x:%02x:%02x:%02x:%02x, mtu=%d, state=%d, features=0x%X, ifindex=0x%X, "
			"flags=0x%X, type=%d, hard_header_len=%d, addr_len=%d",
			__FUNC__,
			dev->name,
			dev,
			dev->dev_addr[0],
			dev->dev_addr[1],
			dev->dev_addr[2],
			dev->dev_addr[3],
			dev->dev_addr[4],
			dev->dev_addr[5],			
			dev->mtu,
			dev->state,
			dev->features,
			dev->ifindex,			
			dev->flags,
			dev->type,
			dev->hard_header_len,			
			dev->addr_len					
			);
	}

	read_unlock(&dev_base_lock);
}

/**
 *	netdev_prepare
 */
#undef __FUNC__
#define __FUNC__ "netdev_prepare()"
static struct sk_buff*
netdev_prepare(
	struct net_device *dev
	)
{
	struct sk_buff *skb = NULL;

	//
	//	We dont want to see loopback ifaces
	//
	if (dev->flags & IFF_LOOPBACK) {
		TRACE(LOG_WARN, "%s: '%s' is loopback device.", __FUNC__, dev->name);
		return NULL;
	}

	//
	//	We dont want to see ifaces with another type then ethernet or ppp
	//
	if (dev->type != ARPHRD_ETHER && dev->type != ARPHRD_PPP) {
		TRACE(LOG_WARN, "%s: '%s' isn't ethernet or ppp device. (%d)", __FUNC__, dev->name, dev->type);
		return NULL;
	}

	skb = dev_alloc_skb(0);

	if (!IS_PTR_VALID(skb)) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate memory for dev='%s'.", __FUNC__, dev->name);
		return NULL;
	}

	(GET_CB(skb))->nf_data = NULL;			
	(GET_CB(skb))->copied_len = 0;
	(GET_CB(skb))->type = NET_DEV_PACKET;
	(GET_CB(skb))->received = false;
	(GET_CB(skb))->policy_routing = false;	

	skb->dev = alloc_nonpaged_mem(sizeof(struct net_device));

	if (!IS_PTR_VALID(skb->dev)) {
		TRACE(LOG_WARN, "%s: alloc_nonpaged_mem(): can't allocate memory for copy of dev='%s'.", __FUNC__, dev->name);
		kfree_skb(skb);
		return NULL;
	}

	memset(skb->dev, 0, sizeof(struct net_device));

	//
	//	save info
	//
	skb->dev->flags = dev->flags;
	skb->dev->mtu = dev->mtu;
	skb->dev->ifindex = dev->ifindex;
	skb->dev->state = dev->state;
	skb->dev->features = dev->features;
	skb->dev->type = dev->type;
	skb->dev->hard_header_len = dev->hard_header_len;			
	skb->dev->addr_len = dev->addr_len;			

	//
	//	little hack: master serves as storage for original dev (adapter id)
	//
	skb->dev->master = dev;

	memcpy(skb->dev->dev_addr, dev->dev_addr, sizeof(dev->dev_addr));
	memcpy(skb->dev->name, dev->name, sizeof(dev->name));

	skb->len = 0;

	return skb;
}

/**
 * netdev_up
 */
#undef __FUNC__
#define __FUNC__ "netdev_up()"
void
netdev_up(
	struct net_device *dev
	)
{
	struct sk_buff *skb = NULL;
	int err = 0;

	skb = netdev_prepare(dev);

	if (!IS_PTR_VALID(skb)) {
		return;
	}

	netdev_dump_skb(skb, __FUNC__);

	if ((err = queue_insert(&g_main_queue, skb, 1))) {

		//
		//	It's normal that queue_insert() will fail after module load, because queue is not enabled.		
		//
		TRACE(LOG_WARN, "%s: queue_insert(): dev='%s', err=%d", __FUNC__, dev->name, err);					

		if (skb->dev) {
			free_nonpaged_mem(skb->dev, sizeof(struct net_device));
			skb->dev = NULL;
		}

		kfree_skb(skb);

		return;
	}

	io_dev_wakeup();
}

/**
 * netdev_down
 */
#undef __FUNC__
#define __FUNC__ "netdev_down()"
void
netdev_down(
	struct net_device *dev
	)
{
	struct sk_buff *skb = NULL;
	int err = 0;

	skb = netdev_prepare(dev);

	if (!IS_PTR_VALID(skb)) {
		return;
	}

	netdev_dump_skb(skb, __FUNC__);

	if ((err = queue_insert(&g_main_queue, skb, 1))) {

		//
		//	It's normal that queue_insert() will fail after module load, because queue is not enabled.		
		//
		TRACE(LOG_WARN, "%s: queue_insert(): dev='%s', err=%d", __FUNC__, dev->name, err);					

		if (skb->dev) {
			free_nonpaged_mem(skb->dev, sizeof(struct net_device));
			skb->dev = NULL;
		}

		kfree_skb(skb);

		return;
	}

	io_dev_wakeup();
}

/**
 * netdev_signal_end
 */
#undef __FUNC__
#define __FUNC__ "netdev_signal_end()"
void
netdev_signal_end(
	void
	)
{
	struct sk_buff *skb = NULL;
	int err = 0;

	skb = dev_alloc_skb(0);

	if (!IS_PTR_VALID(skb)) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate memory.", __FUNC__);
		return;
	}

	(GET_CB(skb))->nf_data = NULL;			
	(GET_CB(skb))->copied_len = 0;
	(GET_CB(skb))->type = NET_DEV_PACKET;
	(GET_CB(skb))->received = false;
	(GET_CB(skb))->policy_routing = false;	

	skb->dev = NULL;	
	skb->len = 0;

	if ((err = queue_insert(&g_main_queue, skb, 1))) {
		TRACE(LOG_WARN, "%s: queue_insert(): err=%d", __FUNC__, err);							
		kfree_skb(skb);
		return;
	}

	io_dev_wakeup();
}

/**
 *	netdev_signal_all_up
 */
#undef __FUNC__
#define __FUNC__ "netdev_signal_all_up()"
void
netdev_signal_all_up(
	void
	)
{
	struct net_device *dev = NULL;
	
	read_lock(&dev_base_lock);
	
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	for_each_netdev(dev) {	
#else
	for_each_netdev(&init_net, dev) {	
#endif	

		//
		//	We dont want to see down ifaces
		//
		if (!(dev->flags & IFF_UP)) {
			continue;
		}

		netdev_up(dev);
	}

	netdev_signal_end();

	read_unlock(&dev_base_lock);
}

/**
 *	netdev_close
 */
#undef __FUNC__
#define __FUNC__ "netdev_close()"
int
netdev_close(
	void
	)
{
	int err = 0;

	if (gs_netdev_notifier_registered) {

		if ((err = unregister_netdevice_notifier(&gs_ipf_netdev_notifier))) {
			TRACE(LOG_WARN, "%s: unregister_netdevice_notifier(): err=%d", __FUNC__, err);			
		} else {	
			TRACE(LOG_INFO, "%s: net device notifier unregistered.", __FUNC__);
		}

		gs_netdev_notifier_registered = 0;
	}

	if (gs_inet_notifier_registered) {

		if ((err = unregister_inetaddr_notifier(&gs_ipf_inet_notifier))) {
			TRACE(LOG_WARN, "%s: unregister_inetaddr_notifier(): err=%d", __FUNC__, err);			
		} else {	
			TRACE(LOG_INFO, "%s: inet addr notifier unregistered.", __FUNC__);
		}

		gs_inet_notifier_registered = 0;
	}

	return 0;
}

/**
 *	netdev_init
 */
#undef __FUNC__
#define __FUNC__ "netdev_init()"
int
netdev_init(
	void
	)
{
	int err = 0;

	//
	//	Register net devices notifier
	//
	if ((err = register_netdevice_notifier(&gs_ipf_netdev_notifier))) {
		TRACE(LOG_ERR, "%s: register_netdevice_notifier(): err=%d", __FUNC__, err);
		return err;
	}
	
	TRACE(LOG_INFO, "%s: net device notifier registered.", __FUNC__);

	gs_netdev_notifier_registered = 1;

	//
	//	Register IP address change notifier
	//
	if ((err = register_inetaddr_notifier(&gs_ipf_inet_notifier))) {
		TRACE(LOG_ERR, "%s: register_inetaddr_notifier(): err=%d", __FUNC__, err);
		return err;
	}

	TRACE(LOG_INFO, "%s: inet addr notifier registered.", __FUNC__);	

	gs_inet_notifier_registered = 1;

	return err;
}

