/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"
#include "socket.h"
#include "pkt.h"

#include <linux/inetdevice.h>

/**
 *	pkt_route
 */
#undef __FUNC__
#define __FUNC__ "pkt_route()"
static inline int 
pkt_route(
	struct sk_buff *skb,
	u_int32_t next_hop
	)
{
	int ret = 0;
	u_int32_t old_dst_ip = ip_hdr(skb)->daddr;

	ip_hdr(skb)->daddr = ((next_hop) ? (next_hop) : (ip_hdr(skb)->daddr));

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	ret = ip_route_me_harder(&skb, RTN_LOCAL);
#else
	ret = ip_route_me_harder(skb, RTN_LOCAL);
#endif
	
	if (ret) {
		TRACE(LOG_WARN, "%s: ip_route_me_harder(): %d", __FUNC__, ret);
		return ret;
	}

	if (!IS_PTR_VALID(skb->dst)
		||
		!IS_PTR_VALID(skb->dst->dev))
	{
		TRACE(LOG_WARN, "%s: no destination or no device specified. dst=%p, dev=%p",
			__FUNC__, skb->dst, ((skb->dst) ? (skb->dst->dev) : (NULL)));
		return -EINVAL;
	}

	ip_hdr(skb)->daddr = old_dst_ip;

#if 0
	ip_hdr(skb)->check = 0;
	ip_hdr(skb)->check	= ip_fast_csum((unsigned char *)ip_hdr(skb), ip_hdr(skb)->ihl);

	skb->ip_summed = CHECKSUM_COMPLETE;
#endif

	skb->dev = skb->dst->dev;

	return 0;
}

/**
 *	ip_finish_output2
 */
#undef __FUNC__
#define __FUNC__ "ip_finish_output2()"
static int32_t 
ip_finish_output2(
	struct sk_buff *skb
	)
{
	int32_t ret = 0;
	struct dst_entry *dst = skb->dst;	
	struct net_device *dev = dst->dev;
	unsigned int hh_len = LL_RESERVED_SPACE(dev);

	skb->protocol = htons(ETH_P_IP);

	skb->mac_header = NULL;
	skb->mac_len = 0;

	/* Be paranoid, rather than too clever. */
#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	if (unlikely(skb_headroom(skb) < hh_len && dev->hard_header)) {
#else
	if (unlikely(skb_headroom(skb) < hh_len && dev->header_ops)) {
#endif	
		struct sk_buff *skb2;

		skb2 = skb_realloc_headroom(skb, LL_RESERVED_SPACE(dev));
		if (skb2 == NULL) {
			kfree_skb(skb);
			return -ENOMEM;
		}
		if (skb->sk)
			skb_set_owner_w(skb2, skb->sk);
		kfree_skb(skb);
		skb = skb2;
	}

	if (dst->hh) {		
 		ret = neigh_hh_output(dst->hh, skb);
		if (ret) {
			TRACE(LOG_WARN, "%s: neigh_hh_output(): %d", __FUNC__, ret);
		}
		return ret;
	}
	else if (dst->neighbour) {		
 		ret = dst->neighbour->output(skb);
		if (ret) {
			TRACE(LOG_WARN, "%s: dst->neighbour->output(): %d", __FUNC__, ret);
		}
		return ret;
	}
	
	TRACE(LOG_WARN, "%s: No header cache and no neighbour !", __FUNC__);

	kfree_skb(skb);
	return -EINVAL;
}

static inline int ip_skb_dst_mtu(struct sk_buff *skb)
{
	struct inet_sock *inet = skb->sk ? inet_sk(skb->sk) : NULL;

	return (inet && inet->pmtudisc == IP_PMTUDISC_PROBE) ?
	       skb->dst->dev->mtu : dst_mtu(skb->dst);
}

int ip_finish_output(struct sk_buff *skb)
{
#if defined(CONFIG_NETFILTER) && defined(CONFIG_XFRM)
	/* Policy lookup after SNAT yielded a new policy */
	if (skb->dst->xfrm != NULL) {
		IPCB(skb)->flags |= IPSKB_REROUTED;
		return dst_output(skb);
	}
#endif
	// GSO: Generic Segmentation Offload 
	if (skb->len > ip_skb_dst_mtu(skb) && !skb_is_gso(skb))
		return ip_fragment(skb, ip_finish_output2);
	else
		return ip_finish_output2(skb);
}

/**
 *	pkt_policy_routing
 */
#undef __FUNC__
#define __FUNC__ "pkt_policy_routing()"
static int32_t 
pkt_policy_routing(
	struct sk_buff *skb,
	unsigned int ip,
	unsigned int ifindex,
	unsigned int *rt_ip
	)
{
	int ret = 0;

	*rt_ip = 0;

	/* policy routing by IP */
	if (ip) {
		*rt_ip = ip;
	}
	/* policy routing by ifindex */
	else if (ifindex) {
		int found = 0;
		struct in_device *in_dev = NULL;
		struct in_ifaddr *ifa = NULL;
		struct in_ifaddr **ifap = NULL;
		struct net_device *dev = netdev_find_by_ifindex(ifindex);

		if (!IS_PTR_VALID(dev)) {
			TRACE(LOG_WARN, "%s: netdev_find_by_ifindex(): can't find netdev (ifindex %d).", __FUNC__, ifindex);
			return -EINVAL;
		}

		if (dev->type != ARPHRD_PPP) {
			TRACE(LOG_WARN, "%s: netdev_find_by_ifindex(): netdev is not a PPP (ifindex %d).", __FUNC__, ifindex);
			return -EINVAL;				
		}

		/* in real just return (struct in_device *)dev->ip_ptr */
		if (!(in_dev = __in_dev_get_rtnl(dev))) {
			TRACE(LOG_WARN, "%s: __in_dev_get_rtnl(): can't get in device for netdev (ifindex %d).", __FUNC__, ifindex);
			return -EINVAL;				
		}

		/* ToDo: just take first destination IP ? */
		for (ifap = &in_dev->ifa_list; (ifa = *ifap) != NULL; ifap = &ifa->ifa_next) {
			if (!strcmp(dev->name, ifa->ifa_label)) {				
				*rt_ip = ifa->ifa_address;
				found = 1;
				break;
			}
		}

		if (!found) {
			TRACE(LOG_WARN, "%s: can't get destination IP for netdev (ifindex %d).", __FUNC__, ifindex);
			return -EINVAL;				
		}		
	}
	/* no policy routing */
	else {		
		return -EINVAL;
	}

	/* ToDo: skb->dst can be NULL
 	 * - called from pkt_send_out()
	 * - called on received packet, skb->dst is filled in ip_rcv_finish() with ip_route_input()
  	 */

	/* re-route packet */
	if ((ret = pkt_route(skb, *rt_ip))) {			
		TRACE(LOG_WARN, "%s: pkt_route(): %d", __FUNC__, ret);		
		return ret;
	}
	
	return 0;
}

/**
 *	pkt_accept
 */
#undef __FUNC__
#define __FUNC__ "pkt_accept()"
int32_t
pkt_accept(	
	struct kipf_packet *packet,
	struct kipf_action *action
	)
{
	struct sk_buff *skb = NULL;			
	int ret = 0;
	int verdict = NF_DROP;

	do {

		if (!IS_PTR_VALID(packet) || !IS_PTR_VALID(action)) {	
			TRACE(LOG_WARN, "%s: invalid parameter(s)", __FUNC__);
			break;
		}	

		if (!(skb = queue_remove(&g_pending_queue, (void*)(u_int32_t) action->id, queue_cmp_by_id))) {		
			TRACE(LOG_WARN, "%s: queue_remove(): can't remove skb with id=%p from pending queue.", __FUNC__, (void*)(u_int32_t) action->id);
			break;
		}				

		if ((GET_CB(skb))->type != NETWORK_PACKET) {
			TRACE(LOG_WARN, "%s: not a network packet ?", __FUNC__);
			break;
		}

		//
		//	Packet was modified in user space ?
		//
		if (action->modif_len > 0) {
			u_int32_t frame_len = 0;
#if defined (USE_SCATTER_GATHER)
			void *kbuffer = NULL;			
#endif
			if (get_user(frame_len, (u_int32_t __user *) &packet->len)) {
				TRACE(LOG_WARN, "%s: get_user(): segmentation fault. skb=%p", __FUNC__, skb);				
				break;
			}

			if (frame_len != sizeof(struct ethhdr) + skb->len) {
				TRACE(LOG_WARN, "%s: frame length was modified. skb=%p", __FUNC__, skb);				
				break;
			}

			if (action->modif_len + sizeof(struct ethhdr) > frame_len) {
				TRACE(LOG_WARN, "%s: modified length is greater than frame length. skb=%p", __FUNC__, skb);				
				break;
			}

#if defined (USE_SCATTER_GATHER)

			//
			//	ToDo: grab memory buffers from preallocated cache
			//
			kbuffer = alloc_paged_mem(action->modif_len);

			if (!IS_PTR_VALID(kbuffer)) {
				TRACE(LOG_WARN, "%s: alloc_paged_mem(): can't allocate %d bytes of memory.", __FUNC__, action->modif_len);				
				break;
			}

			//
			//	1. copy from user space buffer into kernel memory
			//
			if ((ret = copy_from_user(kbuffer, (u_int8_t __user *) packet->data + sizeof(struct ethhdr), action->modif_len))) {					
				TRACE(LOG_WARN, "%s: copy_from_user(): segmentation fault.", __FUNC__);
				free_paged_mem(kbuffer, action->modif_len);				
				break;
			}

			//
			//	2. copy from kernel memory into skb
			//		 			
			if ((ret = skb_store_bits(skb, 0, kbuffer, action->modif_len))) {
				TRACE(LOG_WARN, "%s: skb_copy_bits(): skb=%p, modif_len=%d, ret=%d", __FUNC__, skb, action->modif_len, ret);
				free_paged_mem(kbuffer, action->modif_len);
				break;
			}

			free_paged_mem(kbuffer, action->modif_len);

#else // USE_SCATTER_GATHER

			ip_frag_update_packet_data(skb, packet->data + sizeof(struct ethhdr), action->modif_len);

#endif // USE_SCATTER_GATHER

			TRACE(LOG_DBG, "%s: skb=%p was modified with %d byte(s).",__FUNC__, skb, action->modif_len);
		}

		/* apply policy routing ? */
		if (action->route_info.next_hop || action->route_info.ifindex) {			
			unsigned int rt_ip = 0;			

			if ((ret = pkt_policy_routing(skb, action->route_info.next_hop, action->route_info.ifindex, &rt_ip))) {
				TRACE(LOG_WARN, "%s: pkt_policy_routing(): %d, skb=%p", __FUNC__, ret, skb);
				break;
			}						

			if (IS_DEBUG_FLAG_SET()) {
				if (action->route_info.ifindex) {
					TRACE(LOG_DBG, "%s: skb=%p rerouted throught '%s', remote ip %u.%u.%u.%u", 
					__FUNC__, skb, skb->dev->name, IP_ADDRESS_STRING(rt_ip));		
				}

				if (action->route_info.next_hop) {
					TRACE(LOG_DBG, "%s: skb=%p rerouted throught %u.%u.%u.%u", 
					__FUNC__, skb, IP_ADDRESS_STRING(rt_ip));		
				}								
			}

			(GET_CB(skb))->policy_routing = true;

			/* received packets */
			if ((GET_CB(skb))->received) {			

				/* we want to see received packets as sent */
				(GET_CB(skb))->received = false;

				if ((ret = queue_insert(&g_main_queue, skb, 1))) {				
					TRACE(LOG_WARN, "%s: queue_insert(): can't insert skb=%p into main queue.", __FUNC__, skb);
					break;
				}

				/* avoid invocation of ip_filter_issue_verdict */
				skb = NULL;				
				break;
			} 
		}

		verdict = NF_ACCEPT;

	} while(0);

	if (IS_PTR_VALID(skb)) {
		ip_filter_issue_verdict(skb, verdict);
	}
		
	return ret;
}

/**
 *	pkt_drop
 */
#undef __FUNC__
#define __FUNC__ "pkt_drop()"
int32_t
pkt_drop(	
	struct kipf_packet *packet,
	struct kipf_action *action	
	)
{
	struct sk_buff *skb = NULL;	

	if (!IS_PTR_VALID(action)) {	
		TRACE(LOG_WARN, "%s: invalid parameter(s)", __FUNC__);
		return -EINVAL;
	}

	if (!(skb = queue_remove(&g_pending_queue, (void*)(u_int32_t) action->id, queue_cmp_by_id))) {		
		TRACE(LOG_WARN, "%s: queue_remove(): can't remove skb with id=%p from pending queue.", __FUNC__, (void*)(u_int32_t) action->id);
		return -EINVAL;
	}				

	ip_filter_issue_verdict(skb, NF_DROP);
		
	return 0;   
}

/**
 *	pkt_complete
 */
#undef __FUNC__
#define __FUNC__ "pkt_complete()"
int32_t
pkt_complete(	
	struct kipf_packet *packet,
	struct kipf_action *action
	)
{
	unsigned int length = 0;	
	struct sk_buff *skb = NULL;			
	int frame_len = 0;	
	int bytes_up = 0;
	int32_t ret = -EINVAL;
#if defined (USE_SCATTER_GATHER)
	void *kbuffer = NULL;
	int offset_in_skb = 0;
#endif

	do {

		if (!IS_PTR_VALID(packet) || !IS_PTR_VALID(action)) {	
			TRACE(LOG_WARN, "%s: invalid parameter(s)", __FUNC__);
			break;
		}	

		if (!(skb = queue_remove(&g_pending_queue, (void*)(u_int32_t) action->id, queue_cmp_by_id))) {
			TRACE(LOG_WARN, "%s: queue_remove(): can't remove skb with id=%p from pending queue.", __FUNC__, (void*)(u_int32_t) action->id);
			ret = -EINVAL;
			break;
		}						

		if (get_user(length, (unsigned int __user *) &packet->header.len)) {
			TRACE(LOG_WARN, "%s: get_user(): segmentation fault. skb=%p", __FUNC__, skb);
			ret = -EFAULT;
			break;
		}

		frame_len = sizeof(struct ethhdr) + skb->len;	

		if (length < sizeof(struct kipf_packet) + frame_len) {
			TRACE(LOG_WARN, "%s: buffer is too small. skb=%p, length=%d, frame_len=%d", __FUNC__, skb, length, frame_len);		
			ret = -EINVAL;
			break;
		}

#if defined (USE_SCATTER_GATHER)

		offset_in_skb = (GET_CB(skb))->copied_len - sizeof(struct ethhdr);

		bytes_up = skb->len - offset_in_skb;

		//
		//	ToDo: grab memory buffers from preallocated cache
		//
		kbuffer = alloc_paged_mem(bytes_up);

		if (!IS_PTR_VALID(kbuffer)) {
			TRACE(LOG_WARN, "%s: alloc_paged_mem(): can't allocate %d bytes of memory.", __FUNC__, bytes_up);
			ret = -ENOMEM;
			break;
		}		

		//
		//	copy from skb into kernel memory first
		//		 			
		if ((ret = skb_copy_bits(skb, offset_in_skb, kbuffer, bytes_up))) {
			TRACE(LOG_WARN, "%s: skb_copy_bits(): skb=%p, offset_in_skb=%d, bytes_up=%d, ret=%d", __FUNC__, skb, offset_in_skb, bytes_up, ret);
			free_paged_mem(kbuffer, bytes_up);
			break;
		}

		//
		//	and then from kernel memory into user space buffer
		//
		if (copy_to_user(packet->data + (GET_CB(skb))->copied_len, kbuffer, bytes_up)) {
			free_paged_mem(kbuffer, bytes_up);
			ret = -EFAULT;
			break;
		}

		free_paged_mem(kbuffer, bytes_up);

		(GET_CB(skb))->copied_len += bytes_up;

#else // USE_SCATTER_GATHER

		bytes_up = ip_frag_copy_packet_data(
				skb,
				packet->data + sizeof(struct ethhdr),
				length - sizeof(struct kipf_packet) - sizeof(struct ethhdr),
				(GET_CB(skb))->copied_len - sizeof(struct ethhdr)
				);

		(GET_CB(skb))->copied_len = sizeof(struct ethhdr) + bytes_up;

#endif // USE_SCATTER_GATHER

		if (put_user((GET_CB(skb))->copied_len, &packet->data_len)) {
			TRACE(LOG_WARN, "%s: put_user(): segmentation fault. skb=%p", __FUNC__);
			ret = -EFAULT;
			break;
		}

		TRACE(LOG_DBG, "%s: skb=%p completed. bytes_up=%d, copied_len=%d", __FUNC__, skb, bytes_up, (GET_CB(skb))->copied_len);

		if ((ret = queue_insert(&g_pending_queue, skb, 0))) {
			TRACE(LOG_WARN, "%s: queue_insert(): can't insert skb=%p back into pending queue.", __FUNC__, skb);
			break;
		}

		ret = 0;

	} while(0);

	//
	//	some error ?
	//
	if (ret && IS_PTR_VALID(skb)) {
		ip_filter_issue_verdict(skb, NF_DROP);
	}

	return ret;
}

/**
 *	pkt_send_out
 */
#undef __FUNC__
#define __FUNC__ "pkt_send_out()"
int32_t
pkt_send_out(	
	unsigned char *pkt,
	unsigned int pkt_len,		
	struct kipf_action *action
	)
{	
	struct sk_buff *skb = NULL;
	u_int32_t ret = 0;		
	unsigned int rt_ip = 0;

	if (!IS_PTR_VALID(pkt) || !IS_PTR_VALID(action)) {	
		TRACE(LOG_WARN, "%s: invalid parameter(s)", __FUNC__);
		return -EINVAL;
	}

	if (!pkt_len || pkt_len < sizeof(struct ethhdr) + sizeof(struct iphdr)) {
		TRACE(LOG_WARN, "%s: pkt len is too short. (%d)", __FUNC__, pkt_len);
		return -EINVAL;
	}

	//
	//	dev_alloc_skb() will allocate additional NET_SKB_PAD (16) bytes for headroom
	//
	skb = dev_alloc_skb(pkt_len + NET_IP_ALIGN);

	if (!IS_PTR_VALID(skb)) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate %d bytes of memory.", __FUNC__, pkt_len);
		return -ENOMEM;
	}

	//
	//	increase headroom
	//
	skb_reserve(skb, NET_IP_ALIGN);	
	skb_reset_network_header(skb);

	skb->protocol = htons(ETH_P_IP);

  	if ((ret = copy_from_user(skb_put(skb, pkt_len - sizeof(struct ethhdr)), pkt + sizeof(struct ethhdr), pkt_len - sizeof(struct ethhdr)))) {
		TRACE(LOG_WARN, "%s: copy_from_user(): segmentation fault.", __FUNC__);
		kfree_skb(skb);
		return -EFAULT;
	}	

	if (action->route_info.ifindex) {
		action->route_info.next_hop = 0;
	} else {
		action->route_info.next_hop = ((action->route_info.next_hop) ? (action->route_info.next_hop) : (ip_hdr(skb)->daddr));
	}

	if ((ret = pkt_policy_routing(skb, action->route_info.next_hop, action->route_info.ifindex, &rt_ip))) {
		TRACE(LOG_WARN, "%s: pkt_policy_routing(): %d", __FUNC__, ret);
		kfree_skb(skb);
		return -EINVAL;
	}
	
	if ((ret = ip_finish_output(skb))) {	
		TRACE(LOG_WARN, "%s: ip_finish_output(): %d", __FUNC__, ret);			
		return ret;
	}	

#if 0
	ip_hdr(skb)->saddr = 0x810110ac;
	ip_hdr(skb)->daddr = 0x800110ac;

	dump_ip_packet((u_int8_t*) ip_hdr(skb), skb->len);
#endif

	return 0;
}

/**
 *	pkt_send_in
 */
#undef __FUNC__
#define __FUNC__ "pkt_send_in()"
int32_t
pkt_send_in(	
	unsigned char *pkt,
	unsigned int pkt_len,	
	struct kipf_action *action
	)
{
	struct sk_buff *skb = NULL;		
	u_int8_t *dst = NULL;

	if (!IS_PTR_VALID(pkt) || !IS_PTR_VALID(action)) {	
		TRACE(LOG_WARN, "%s: invalid parameter(s)", __FUNC__);
		return -EINVAL;
	}

	if (!pkt_len || pkt_len < sizeof(struct ethhdr) + sizeof(struct iphdr)) {
		TRACE(LOG_WARN, "%s: pkt len is too short. (%d)", __FUNC__, pkt_len);
		return -EINVAL;
	}

	skb = dev_alloc_skb(pkt_len + NET_IP_ALIGN);

	if (!IS_PTR_VALID(skb)) {
		TRACE(LOG_WARN, "%s: dev_alloc_skb(): can't allocate %d+%d bytes of memory.", __FUNC__, pkt_len, NET_IP_ALIGN);
		return -ENOMEM;
	}

	skb_reserve(skb, NET_IP_ALIGN);		

	dst = skb_put(skb, pkt_len);

	if (copy_from_user(dst, (u_int8_t __user *)pkt, pkt_len)) {
		TRACE(LOG_WARN, "%s: copy_from_user(): segmentation fault.", __FUNC__);
		kfree_skb(skb);
		return -EFAULT;
	}

	if (action->route_info.ifindex) {
		skb->dev = netdev_find_by_ifindex(action->route_info.ifindex);
	} else if (action->route_info.next_hop) {
		//	not implemented
		skb->dev = NULL;
	} else {
		//	last chance
		skb->dev = netdev_find_by_mac(((struct ethhdr*)dst)->h_dest, ETH_ALEN);
	}

	if (!IS_PTR_VALID(skb->dev)) {
		TRACE(LOG_WARN, "%s: netdev_find(): can't find netdev. (ifindex=%d, next_hop=%d)",
			__FUNC__,
			action->route_info.ifindex,
			action->route_info.next_hop
			);
		kfree_skb(skb);
		return -EINVAL;
	}

	//
	//	loopback device ?
	//
	if (skb->dev->flags & IFF_LOOPBACK) {	
		TRACE(LOG_WARN, "%s: packet handling on loopback device isn't permitted. dev='%s'", __FUNC__, skb->dev->name);
		kfree_skb(skb);
		return -EINVAL;
	}

	//
	//	non ethernet or ppp device ?
	//
	if (skb->dev->type != ARPHRD_ETHER && skb->dev->type != ARPHRD_PPP) {
		TRACE(LOG_WARN, "%s: '%s' isn't ethernet or ppp device. (%d)", __FUNC__, skb->dev->name, skb->dev->type);
		kfree_skb(skb);
		return -EINVAL;		
	}

	skb->protocol = eth_type_trans(skb, skb->dev);
	skb->mac_len = sizeof(struct ethhdr);

	//
	//	we don't want to see indicated packet again
	//
	skb->mark = PKT_SOCKET_TAG;

	//skb->ip_summed = CHECKSUM_UNNECESSARY;

	TRACE(LOG_DBG, "%s: skb=%p will be indicated to tcpip via '%s'.", __FUNC__, skb, skb->dev->name);

	//
	//	Receiving packets in non-interrupt mode:
	//	should use netif_rx_ni (it will call preempt_disable > netif_rx -> preempt_enable
	//
	netif_rx_ni(skb);

	return 0;
}

