/*
 * Kerio IP filter
 * Copyright (C) Kerio Technologies s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "hdrs.h"

/**
 *	Locals
 */
struct proc_dir_entry *gs_debug_flag_ent = NULL;
struct proc_dir_entry *gs_status_ent = NULL;
struct proc_dir_entry *gs_new_dir_ent = NULL;

unsigned char gs_debug_flag[2];

/**
 *	proc_read_status
 */
#undef __FUNC__
#define __FUNC__ "proc_read_status()"
static int 
proc_read_status(
	char *buf,
	char **start,
	off_t pos,
	int count,
	int *eof,
	void *data
	)
{	
	if (count < 256) {
		TRACE(LOG_WARN, "%s: invalid len. (%d)", __FUNC__, count);
		return -EINVAL;
	}

	return sprintf(buf,
			"device: %s\n"
			"main queue len: %d\n"
			"pending queue len: %d\n",
			((g_device_opened) ? ("opened") : ("closed")),
			queue_len(&g_main_queue),
			queue_len(&g_pending_queue)
			);
}

/**
 *	proc_read_debug_flag
 */
#undef __FUNC__
#define __FUNC__ "proc_read_debug_flag()"
static int 
proc_read_debug_flag(
	char *buf,
	char **start,
	off_t pos,
	int count,
	int *eof,
	void *data
	)
{	
	int len = sizeof(gs_debug_flag);

	if (count < len) {
		TRACE(LOG_WARN, "%s: invalid len. (%d)", __FUNC__, count);
		return -EINVAL;
	}

	memcpy(buf, gs_debug_flag, len);

	TRACE(LOG_INFO, "%s: debug_flag='%c'", __FUNC__, gs_debug_flag[0]);

	return len;
}

/**
 *	proc_write_debug_flag
 */
#undef __FUNC__
#define __FUNC__ "proc_write_debug_flag()"
static int 
proc_write_debug_flag(
	struct file *file,
	const char __user * buf,
	unsigned long count,
	void *data
	)
{	
	int len = sizeof(gs_debug_flag);			

	if (count > len) {
		TRACE(LOG_WARN, "%s: invalid len. (%d)", __FUNC__, count);
		return -EINVAL;
	}

	if (copy_from_user(gs_debug_flag, buf, count)) {
		TRACE(LOG_WARN, "%s: get_user(): segmentation fault.", __FUNC__);
		return -EFAULT;			
	}		

	gs_debug_flag[len-1] = '\n';

	//
	//	Set global debug variable
	//
	g_utils_debug_enabled = ((gs_debug_flag[0] == '1') ? (1) : (0));

	TRACE(LOG_INFO, "%s: debug_flag='%c'", __FUNC__, gs_debug_flag[0]);

	return count;		
}

/**
 *	proc_init
 */
#undef __FUNC__
#define __FUNC__ "proc_init()"
int32_t
proc_init(
	void
	)
{	
	gs_debug_flag[0] = ((g_utils_debug_enabled) ? ('1') : ('0'));
	gs_debug_flag[1] = '\n';

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
	gs_new_dir_ent = proc_mkdir("kipf", proc_net);
#else
	gs_new_dir_ent = proc_mkdir("kipf", init_net.proc_net);
#endif

	if (!gs_new_dir_ent) {
		TRACE(LOG_ERR, "%s: proc_mkdir(): can't create \"/proc/net/kipf\" directory.", __FUNC__);
		return -1;
	}

	TRACE(LOG_INFO, "%s: directory \"/proc/net/kipf\" created.", __FUNC__);

	gs_debug_flag_ent = create_proc_entry(
							"debug_flag",
							S_IWUSR | S_IRUGO,	// write acces only for owner, read acces for owner, group, other
							gs_new_dir_ent
							);

	if (!gs_debug_flag_ent) {
		TRACE(LOG_ERR, "%s: create_proc_entry(): can't create \"/proc/net/kipf/debug_flag\" entry.", __FUNC__);
		return -1;
	}

	gs_debug_flag_ent->read_proc = proc_read_debug_flag;
	gs_debug_flag_ent->write_proc = proc_write_debug_flag;
	gs_debug_flag_ent->data = (void *) 0x12345678;

	TRACE(LOG_INFO, "%s: file \"/proc/net/kipf/debug_flag\" created.", __FUNC__);

	gs_status_ent = create_proc_entry(
						"status",
						S_IRUGO,	// read acces for owner, group, other
						gs_new_dir_ent
						);

	if (!gs_status_ent) {
		TRACE(LOG_ERR, "%s: create_proc_entry(): can't create \"/proc/net/kipf/status\" entry.", __FUNC__);
		return -1;
	}

	gs_status_ent->read_proc = proc_read_status;
	gs_status_ent->data = (void *) 0x87654321;

	TRACE(LOG_INFO, "%s: file \"/proc/net/kipf/status\" created.", __FUNC__);

	return 0;
}

/**
 *	proc_close
 */
#undef __FUNC__
#define __FUNC__ "proc_close()"
int32_t
proc_close(
	void
	)
{
	if (IS_PTR_VALID(gs_new_dir_ent)) {

		if (IS_PTR_VALID(gs_status_ent)) {

			remove_proc_entry(
				"status",
				gs_new_dir_ent
				);

			TRACE(LOG_INFO, "%s: file \"/proc/net/kipf/status\" removed.", __FUNC__);

			gs_status_ent = NULL;
		}

		if (IS_PTR_VALID(gs_debug_flag_ent)) {

			remove_proc_entry(
				"debug_flag",
				gs_new_dir_ent
				);

			TRACE(LOG_INFO, "%s: file \"/proc/net/kipf/debug_flag\" removed.", __FUNC__);

			gs_debug_flag_ent = NULL;
		}

#if defined(LINUX_KERNEL_2_6_22_OR_2_6_23)
		remove_proc_entry("kipf", proc_net);
#else
		remove_proc_entry("kipf", init_net.proc_net);
#endif

		TRACE(LOG_INFO, "%s: directory \"/proc/net/kipf\" removed.", __FUNC__);

		gs_new_dir_ent = NULL;
	}

	return 0;
}

