/*
 * Kerio IP Filter library for userspace.
 * Copyright (C) Kerio Technologies, s.r.o.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>
#include <pthread.h>
#include <stdarg.h>
#include <wchar.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>

#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/errno.h>

#include <kipf/api.h>

/* Globals */
static int fdesc = -1;

/**
 *	kipf_get_sys_error
 */
unsigned int kipf_get_sys_error(void)
{
	return errno;
}

/*
 *	kipf_open
 */
kipf_status kipf_open(void)
{
	//
	//	check if pointers is 4bytes long (32 bit application)
	//
	if (sizeof(void*) != 4) {
		return KIPF_MAKE_STATUS(KIPF_NO_32BIT_APP);		
	}

	/*	Already opened ? */	
	if (fdesc != -1) {
		return KIPF_MAKE_STATUS(KIPF_ERROR_ALREADY_OPENED);
	}

	if ((fdesc = open(KIPF_IO_DEV, O_RDWR)) == -1) {
		return KIPF_MAKE_STATUS(KIPF_OPEN_DEVICE_ERROR);
	}

	/*	Set to non-blocking mode  */
	if (fcntl(fdesc, F_SETFL, O_NONBLOCK) == -1) {				
		close(fdesc);
		fdesc = -1;
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

/*
 *	kipf_close
 */
kipf_status kipf_close(unsigned int rule)
{
	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_NOT_OPENED);
	}	

	ioctl(fdesc, KIPF_IOCTL_SET_DEFAULT_TRAFFIC_RULE, &rule);

	close(fdesc);
	fdesc = -1;

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

/*
 *	kipf_read
 */
kipf_status kipf_read(READ_REQ *rreq)
{
	int ret = 0;
	fd_set readfds;	
	fd_set exceptfds;
	//struct timeval m_timeout;

	//m_timeout.tv_sec = 0;
	//m_timeout.tv_usec = 0;	

	if (fdesc == -1 || !rreq || rreq->count > MAX_RW_BUFFERS_COUNT) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

try_again:

	FD_ZERO(&readfds);
	FD_ZERO(&exceptfds);
	
	if (fdesc >= FD_SETSIZE)
		return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);

	FD_SET(fdesc, &readfds);
	FD_SET(fdesc, &exceptfds);

	ret = select(fdesc + 1, &readfds, NULL, &exceptfds, NULL /*&m_timeout*/);		

	if (ret > 0) {

		if (FD_ISSET(fdesc, &exceptfds)) {
			return KIPF_MAKE_STATUS(KIPF_READ_CANCELLED);			
		}

		if (FD_ISSET(fdesc, &readfds)) {

			if ((ret = read(fdesc, rreq, sizeof(READ_REQ))) == -1) {	
				return KIPF_MAKE_STATUS(KIPF_READ_ERROR);
			}
	
			rreq->count = ret;

			return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
		}

	} else if (ret == 0) {
		/* timeout ? */
		return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);						
 	} else if (ret == -1) {
		if (errno == EINTR) {			
			goto try_again;
		}
		return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);				
	}			
	
	return KIPF_MAKE_STATUS(KIPF_SELECT_ERROR);		
}

/*
 *	kipf_write
 */
kipf_status kipf_write(WRITE_REQ *wreq)
{
	int ret = 0;

	if (fdesc == -1 || !wreq || wreq->count > MAX_RW_BUFFERS_COUNT) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = write(fdesc, wreq, sizeof(WRITE_REQ))) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_WRITE_ERROR);
	}

	wreq->count = ret;

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

/*
 *	kipf_cancel_read
 */
kipf_status kipf_cancel_read(void)
{
	int ret = 0;

	if (fdesc == -1) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_CANCEL_PENDING_READ, NULL)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);	
}

/*
 *	kipf_get_driver_info
 */
kipf_status kipf_get_driver_info(struct kipf_info *info)
{
	int ret = 0;

	if (fdesc == -1 || !info) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	if ((ret = ioctl(fdesc, KIPF_IOCTL_GET_INFO, info)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}	

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);		
}

/*
 *	kipf_do_ioctl
 */
kipf_status kipf_do_ioctl(struct kipf_packet *packet, unsigned char *pkt, unsigned int pkt_len, unsigned int type, unsigned int modif_len, struct kipf_route_info *route_info)
{
	IOCTL_REQ ioctlRequest;
	int ret = 0;

	if (fdesc == -1 || (!packet && !pkt)) {
		return KIPF_MAKE_STATUS(KIPF_INVALID_PARAMETER);
	}

	ioctlRequest.action.id = packet ? packet->id : 0;
	ioctlRequest.action.modif_len = modif_len;
	ioctlRequest.action.route_info.ifindex = route_info ? route_info->ifindex : 0;
	ioctlRequest.action.route_info.next_hop = route_info ? route_info->next_hop : 0;
	ioctlRequest.action.type = type;
	ioctlRequest.pkt = pkt;
	ioctlRequest.pkt_len = pkt_len;
	ioctlRequest.packet = (KIPF_PTR) (packet ? packet : 0);

	if ((ret = ioctl(fdesc, KIPF_IOCTL_HANDLE_PACKET, &ioctlRequest)) == -1) {	
		return KIPF_MAKE_STATUS(KIPF_IOCTL_ERROR);
	}

	return KIPF_MAKE_STATUS(KIPF_NO_ERROR);
}

/*
 *	kipf_complete_packet
 */
kipf_status kipf_complete_packet(struct kipf_packet *packet)
{
	return kipf_do_ioctl(packet, 0, 0, KIPF_ACTION_COMPLETE, 0, NULL);
}

/*
 *	kipf_drop_packet
 */
kipf_status kipf_drop_packet(struct kipf_packet *packet)
{
	return kipf_do_ioctl(packet, 0, 0, KIPF_ACTION_DROP, 0, NULL);
}

/*
 *	kipf_accept_packet
 */
kipf_status kipf_accept_packet(struct kipf_packet *packet)
{
	return kipf_do_ioctl(packet, 0, 0, KIPF_ACTION_ACCEPT, 0, NULL);
}

/*
 *	kipf_accept_packet_with_info
 */
kipf_status kipf_accept_packet_with_info(struct kipf_packet *packet, unsigned int modif_len, struct kipf_route_info *route_info)
{
	return kipf_do_ioctl(packet, 0, 0, KIPF_ACTION_ACCEPT, modif_len, route_info);
}

/*
 *	kipf_send_outgoing_packet
 */
kipf_status kipf_send_outgoing_packet(unsigned char *pkt, unsigned int pkt_len)
{
	struct kipf_route_info route_info = {0, 0};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_OUT, 0, &route_info);	
}

/*
 *	kipf_send_outgoing_packet_by_dev (PPP - ok (eth hdr not needed), ETH - not implemented)
 */
kipf_status kipf_send_outgoing_packet_by_dev(unsigned char *pkt, unsigned int pkt_len, unsigned int ifindex)
{
	struct kipf_route_info route_info = {0, ifindex};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_OUT, 0, &route_info);	
}

/*
 *	kipf_send_outgoing_packet_by_next_hop (PPP - ok (eth hdr not needed), ETH - ok (eth hdr built from params))
 */
kipf_status kipf_send_outgoing_packet_by_next_hop(unsigned char *pkt, unsigned int pkt_len, unsigned int next_hop)
{
	struct kipf_route_info route_info = {next_hop, 0};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_OUT, 0, &route_info);		
}

/*
 *	kipf_send_incoming_packet
 */
kipf_status kipf_send_incoming_packet(unsigned char *pkt, unsigned int pkt_len)
{
	struct kipf_route_info route_info = {0, 0};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_IN, 0, &route_info);		
}

/*
 *	kipf_send_incoming_packet_by_dev (PPP - ok, ETH - ok (use route_pkt()) )
 */
kipf_status kipf_send_incoming_packet_by_dev(unsigned char *pkt, unsigned int pkt_len, unsigned int ifindex)
{
	struct kipf_route_info route_info = {0, ifindex};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_IN, 0, &route_info);		
}

/*
 *	kipf_send_incoming_packet_by_next_hop (PPP - not implemented, ETH - not implemented)
 */
kipf_status kipf_send_incoming_packet_by_next_hop(unsigned char *pkt, unsigned int pkt_len, unsigned int next_hop)
{
	struct kipf_route_info route_info = {next_hop, 0};
	return kipf_do_ioctl(0, pkt, pkt_len, KIPF_ACTION_SEND_IN, 0, &route_info);	
}
